from django.core.cache import caches
from src.django_cache_pydantic.conf import conf


class CacheMetaOptions:
    """
    Helper class for defining options for the cache meta.

    This class assists in defining options for the cache meta class used in PydanticCachedModel.
    """

    default_cache_attr = ['ttl', 'cache_backend', 'primary_key_field', 'verbose']

    def __init__(self, cls, original_cache_class) -> None:
        """
        Initialize the CacheMetaOptions instance with default or user defined values.
        """

        default_dict = {key: None for key in self.default_cache_attr}
        default_dict.update(
            {
                'ttl': conf.CACHE_PYDANTIC_DEFAULT_TTL,
                'cache_backend': caches[conf.CACHE_PYDANTIC_DEFAULT_CACHE],
                'verbose': cls.__name__,
            }
        )
        self.defaults = default_dict
        self.originals = {}
        if original_cache_class is not None:
            self.originals = original_cache_class.__dict__.copy()

    def define_options_class(self):
        """
        Dynamically defines a cache meta class with default and overridden attributes.

        returns: type: The created CacheMeta class.

        """
        original_attrs = {key: value for key, value in self.originals.items() if key in self.default_cache_attr}
        cache_meta_class_attrs = {key: original_attrs.get(key) or self.defaults[key] for key in self.default_cache_attr}
        cache_meta_class = type('CacheMeta', (), cache_meta_class_attrs)
        return cache_meta_class



class Manager:
    """
    Manager class for handling model instances and cache interactions.

    This class provides methods for creating and retrieving pydantic model instances.
    """

    def __init__(self, model_class, cache_meta_class) -> None:
        self.model_class = model_class
        self.cache_meta_class = cache_meta_class

    def create(self, **kwargs):
        """
        Create a new model instance, save it to the cache, and return the instance.
        """

        instance = self.model_class(**kwargs)
        instance.save()
        return instance

    def get(self, pk):
        """
        Retrieve a model instance from the cache based on the primary key.
        """

        cache = self.cache_meta_class.cache_backend
        composite_id = f'{self.model_class.__name__.lower()}__{pk}'
        return cache.get(composite_id)

    def delete(self, pk):
        """
        Delete a model instance from the cache based on the primary key.
        """

        cache = self.cache_meta_class.cache_backend
        composite_id = f'{self.model_class.__name__.lower()}__{pk}'
        return cache.delete(composite_id)


class CacheManager:
    """
    Descriptor class allowing to access objects manager from class not instances.
    """

    manager_class = Manager

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, cls):
        """
        Create a manager for owner class and add it to cache meta class of the owner.
        raises: AttributeError if trying to access the manager from an instance.
        """

        if obj is not None:
            raise AttributeError("Manager for %s isn't accessible via %s instances" % (self.name, cls.__name__))
        cache_manager = getattr(cls._cache_meta, 'manager', None)
        if cache_manager is None:
            cache_manager = self.manager_class(cls, cls._cache_meta)
            setattr(cls._cache_meta, 'manager', cache_manager)
        return cache_manager



from __future__ import annotations

from django.conf import settings
from src.django_cache_pydantic.defaults import default_cache, default_ttl


class DjangoSettings:
    @property
    def CACHE_PYDANTIC_DEFAULT_CACHE(self) -> str:
        return getattr(settings, 'CACHE_PYDANTIC_DEFAULT_CACHE', default_cache)

    @property
    def CACHE_PYDANTIC_DEFAULT_TTL(self) -> int:
        return getattr(settings, 'CACHE_PYDANTIC_DEFAULT_TTL', default_ttl)


conf = DjangoSettings()



from typing import ClassVar
from uuid import uuid4

from pydantic import BaseModel, computed_field

from .manager import CacheManager
from .options import CacheMetaOptions


class PydanticCachedModel(BaseModel):
    """
    A Pydantic model with caching functionality.

    This class extends Pydantic's BaseModel and provides caching features.

    Attributes:
        objects (ClassVar): Model manager with a similar interface to django ORM.
    """

    objects: ClassVar = CacheManager()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs) -> None:
        """
        Initialize the PydanticCachedModel subclass.
        Add CacheMeta class to subclassed children.

        raises: TypeError: If CacheMeta is not a class.

        """
        super().__pydantic_init_subclass__(**kwargs)
        cache_class = getattr(cls, 'CacheMeta', None)
        if cache_class is not None and type(cache_class) is not type:
            raise TypeError('CacheMeta should be a class and not of type %s' % type(cache_class))
        new_cache_class = CacheMetaOptions(cls, cache_class).define_options_class()
        setattr(cls, '_cache_meta', new_cache_class)

    def __init__(self, /, **data):
        super().__init__(**data)
        _internal_id = uuid4().hex
        setattr(self, '_internal_id', _internal_id)

    @computed_field
    def id(self) -> str:
        """
        Computed property for generating the composite ID.

        :return: str: The composite ID.
        """
        pk_field = getattr(self.CacheMeta, 'primary_key_field', None)
        return getattr(self, pk_field) if pk_field is not None else self._internal_id

    def _get_composite_id(self) -> str:
        """
        Returns composite ID composed of class name and internal id.
        :return: str: The composite ID.
        """
        return f'{self.__class__.__name__.lower()}__{self.id}'

    @computed_field
    def pk(self) -> str:
        return self.id

    def save(self):
        """
        Saves the instance in the cache with a specified time-to-live (ttl).
        """
        cache = self._cache_meta.cache_backend
        composite_id = self._get_composite_id()
        cache.set(composite_id, self, self._cache_meta.ttl)

    def delete(self):
        """
        Delete the instance from the cache.
        """
        cache = self._cache_meta.cache_backend
        composite_id = self._get_composite_id()
        cache.delete(composite_id)
